SELECT Path,Name,Size,Attributes,TO_UTCTIME(CreationTime) as CreationTimeUTC,TO_UTCTIME(LastAccessTime) as LastAccessTimeUTC,TO_UTCTIME(LastWriteTime) as LastWriteTimeUTC,FileVersion,ProductVersion,REPLACE_CHR(InternalName, '\u000a\u000d,', ' ')as InternalName,REPLACE_CHR(ProductName, '\u000a\u000d,', ' ')as ProductName,REPLACE_CHR(CompanyName, '\u000a\u000d,', ' ')as CompanyName,REPLACE_CHR(LegalCopyright, '\u000a\u000d,', ' ')as LegalCopyright,REPLACE_CHR(LegalTrademarks, '\u000a\u000d,', ' ')as LegalTrademarks,PrivateBuild,SpecialBuild,REPLACE_CHR(Comments, '\u000a\u000d,', ' ')as Comments,REPLACE_CHR(FileDescription, '\u000a\u000d,', ' ')as FileDescription,OriginalFilename FROM %systemroot%\prefetch\*.*